/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.

   TEXTURES.C - Textures in 256 colors.
   */

/* the includes */
#include "deu.h"
#include "struct_h.h"
#include <assert.h>

typedef struct _resource {
	char *name;
	void *data;
	struct _resource *next;
} resource;

resource *ResourceList;

Bool ForgetAfter = TRUE;
/* whether DisplayPic should de-allocate the picture resource
   after displaying it */

/*
   display a floor or ceiling texture at coords x0, y0 and not beyond x1, y1
   */

#define IMAGE_CONTEXT_SIZE      (((sizeof(GrContext)+15)&~15)+4)

void DisplayFloorTexture( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *texname)
{
	MDirPtr dir;
	
	dir = FindMasterDir( MasterDir, texname);
	if (dir == NULL || dir->dir.size < 4096)
	{
		SetColor( DARKGRAY);
		DrawScreenLine( x0, y0, x1, y1);
		DrawScreenLine( x0, y1, x1, y0);
		return;
	}

	if (dir->dir.size >= 4096 && dir->dir.size < 128*128)
	{
		GrContext gc; // [RH] allocate storage on the stack
		char pixels[4096];
		char *mem[4];
		BasicWadSeek( dir->wadfile, dir->dir.start);
		
		mem[0] = pixels;
		mem[1] = mem[2] = mem[3] = 0;
		GrCreateFrameContext (GR_frameRAM8, 64, 64, mem, &gc);
		BasicWadRead( dir->wadfile, pixels, 4096L);
		GrBitBlt (NULL, x0, y0, &gc, 0, 0, 64, 64, GrWRITE);
	}
	else
	{
		BYTE *pixels, *pstart;
		int dim, step, rowstep, x, y, xp, yp;

		// High-res flats are determined by their size and can be
		// 128x128, 256x256, or 512x512. When rendered, they always
		// get scaled down to cover 64x64 world units.
		if (dir->dir.size < 256*256)
		{
			dim = 128;
		}
		else if (dir->dir.size < 512*512)
		{
			dim = 256;
		}
		else
		{
			dim = 512;
		}
		step = dim / 64;
		rowstep = dim * (step - 1);
		pixels = pstart = (BYTE *)myalloc (dim*dim);
		BasicWadSeek (dir->wadfile, dir->dir.start);
		BasicWadRead (dir->wadfile, pixels, dim*dim);

		xp = x0 + 64;
		yp = y0 + 64;

		if (UseMouse)
			HideMousePointer ();

		// Evil (slow) putpixel time.
		for (y = -64; y != 0; ++y)
		{
			for (x = -64; x != 0; ++x)
			{
				GrPlot (xp+x, yp+y, *pixels);
				pixels += step;
			}
			pixels += rowstep;
		}

		if (UseMouse)
			ShowMousePointer ();

		free (pstart);
	}
}

void GetPicSize(char *picname, BCINT *width, BCINT *height)
{	
	PHeader *h;
	PColumn **columns, *c;
	long p_resource;
	BCINT x; 
	BCINT foo = 0;

	if (picname[0] == ':')
	{
		*width = 32;
		*height = 32;
		return;
	}

	p_resource = (long)GetResource(picname);
	if(!p_resource)
		return;
	
	h = (PHeader *)p_resource;
	columns = (PColumn **)(p_resource + sizeof(PHeader));

	*width = h->width;

	for (x = 0; x < h->width; x = x + 1)
	{
		int top = -1;	// [RH] Tall patch support
		c = (PColumn *)((long)columns[x] + p_resource);
		while(c->yofs != 255)
		{
			if (c->yofs <= top)
			{
				top += c->yofs;
			}
			else
			{
				top = c->yofs;
			}
			if (c->npixels + top > foo)
				foo = c->npixels + top;
			c = (PColumn *)((long)c + (long)(c->npixels) + 4L);
		}
	}

	*height = foo;
}

/*
   display a picture "picname" at coords x0, y0 and not beyond x1, y1
   */


/* SO 2/4/95:
   DisplayPic now uses GetResource to get the sprite/patch to display.
   It checks the variable ForgetAfter to see whether it should
   call ForgetResource after displaying; this should be set by
   the caller.
   ChooseSprite sets this to TRUE; DisplayWallTexture sets it to FALSE.
   DisplayWallTexture then takes the responsibility for calling
   ForgetResource for all the patches in the texture being displayed.
   This enables us to buffer all the patches in case they are used
   more than once in a texture (e.g. LITE5 in Doom (extreme case!)) */

void DisplayPic( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *picname, Bool istexture)
{	
	PHeader *h;
	const PColumn *c;
	long *columns;
	int x, y;
	int xmax;
	BYTE *pixels;
	BYTE *p_resource;

	if(bioskey(1))
	    return;

	if (picname[0] == ':')
	{
		DrawIcon1 (x0, y0, x1, y1, picname+1);
		return;
	}

	p_resource = GetResource(picname);
	if(!p_resource)
		return;

	h = (PHeader *)p_resource;
	columns = (long *)(p_resource + sizeof(PHeader));

	xmax = min (h->width, x1-x0);

	for(x = 0; x < xmax; x = x + 1)
	{
		int top;

		top = -1;
		c = (PColumn *)(columns[x] + p_resource);
		while(c->yofs != 255)
		{
			pixels = (BYTE *)((long)c + (long)sizeof(PColumn) + 1L);
			// [RH] Tall patch support
			if (c->yofs <= top)
			{
				top += c->yofs;
			}
			else
			{
				top = c->yofs;
			}
			for(y = 0; y < c->npixels && (y + top) < (y1 - y0); ++y)
			{
				GrPlot (x + x0, y + y0 + top, *pixels++);
			}
			c = (PColumn *)((long)c + (long)(c->npixels) + 4L);
		}
	}
	/*
	if (Caching == FALSE)
		if(ForgetAfter)
	*/
	ForgetResource(picname);
}

// [RH] Display a texture with fractional scaling
void DisplayPicFrac( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *picname, long xstep, long ystep)
{	
	const PHeader *h;
	const PColumn **columns, *c;
	int x, y;
	BYTE *pixels;
	long p_resource;
	long xp, yp, height;

	if(bioskey(1) && Caching == FALSE)
		return; 

	if ((xstep == 65536 && ystep == 65536) || xstep == 0 || ystep == 0)
	{
		DisplayPic (x0, y0, x1, y1, picname, 1);
		return;
	}

	if (picname[0] == ':')
	{
		DrawIconFrac (x0, y0, x1, y1, picname+1, xstep, ystep);
		return;
	}

	p_resource = (long)GetResource(picname);
	if(!p_resource)
		return;

	h = (const PHeader *)p_resource;
	columns = (const PColumn **)(p_resource + sizeof(PHeader));
	for (x = 0, xp = x0; (x>>16) < h->width && xp < x1; x += xstep, xp++)
	{
		int top;

		top = -1;
		c = (PColumn *)((long)columns[x >> 16] + p_resource);

		while (c->yofs != 255)
		{
			pixels = (BYTE *)((long)c + (long)sizeof(PColumn) + 1L);
			if (c->yofs <= top)
			{
				top += c->yofs;
			}
			else
			{
				top = c->yofs;
			}
			height = c->npixels << 16;
			yp = (top + y0) << 16;
			y = yp % ystep;
			yp /= ystep;
			for (y = 0; y < height && yp < y1; y += ystep, ++yp)
			{
				GrPlot (xp, yp, pixels[y>>16]);
			}
			c = (PColumn *)((long)c + (long)(c->npixels) + 4L);
		} 
	}
}

// [RH] Draw the sprite in the given region, picking a scale in order to keep it inside the box.
void DisplayPicFit( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *picname, Bool istexture)
{	
	long xmax, ymax, xstep, ystep, step;
	long p_resource;
	int width, height;

	if(bioskey(1))
	    return;

	if (picname[0] == ':')
	{
		width = height = 32;
	}
	else
	{
		PHeader *h;
		p_resource = (long)GetResource(picname);
		if(!p_resource)
			return;

		h = (PHeader *)p_resource;
		width = h->width;
		height = h->height;
	}

	xmax = x1 - x0 + 1;
	ymax = y1 - y0;

	xstep = width * 65536 / xmax;
	ystep = height * 65536 / ymax;
	step = max (65536, max (xstep, ystep));

	DisplayPicFrac (x0+(xmax-width*65536/step)/2, y0*step/65536 + (ymax*step-65536*height)/131072, x1, y1, picname, step, step);
}

void DisplaySprite( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *picname)
{
	DisplayPic(x0, y0, x1, y1, picname, 0);
}

void *GetResource(char *name)
{
	void *t = NULL;
	resource *p, *new = (resource *)NULL;
	MDirPtr r;
	
	for(p = ResourceList; p; p = p->next)
		if(!strcmp(name, p->name))
			return p->data;
	
	r = FindMasterDir(MasterDir, name);
	if(r && (t = myalloc(r->dir.size)))
	{
		BasicWadSeek(r->wadfile, r->dir.start);
		BasicWadRead(r->wadfile, t, r->dir.size);
		
		new = (resource *)myalloc(sizeof(resource));
		assert(new);
		new->data = t;
		new->name = strdup(name);
		new->next = ResourceList;
		ResourceList = new;
		return t;
	}
	return NULL;
}

void ForgetResource(char *name)
{
	resource *p, **q = &ResourceList;
	
	for(p = ResourceList; p->next; q = &p->next, p = p->next)
	{
		if(!strcmp(p->name, name))
		{
			*q = p->next;
			
			free(p->data);
			p->data = NULL;
			free(p->name);
			p->name = NULL;
			free(p);
			
			return;
		}
	}
}

void ForgetAllResources()
{
	resource *p, *q;
	for(p = ResourceList; p; p = q)
	{
		q = p->next;
		free(p->data);
		free(p->name);
		free(p);
	}
	ResourceList = (resource *)NULL;
}


Texture *FindTexture(char *name)
{
	Texture *t;
	SList p;
	
	for(p = Texture_sections; p; p = p->next) {
		t = FindTextureIn(name, GetResource(p->string));
		if(t)
			return t;
	}
	return NULL;
}

Texture *FindTextureIn(char *name, void *tblock)
{
	long i, *offsets = tblock;
	Texture *t;
	
	if(tblock)
		for(i = 1; i <= offsets[0]; i++) {
			t = (Texture *)((long)offsets[i] + (long)tblock);
			if(!strnicmp(t->name, name, 8)) /*jff don't care about case*/
				return t;
		}
	
	return NULL;
}

/*
   display a wall texture at coords x0, y0
   */

void DisplayWallTexture( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *texname)
{
	BCINT	n, xsize, ysize;
	Texture  *tinfo;
	PatchDesc *pinfo;
	char	 picname[9];
	long patchname;
	long scalex, scaley;
	void *Pnames;
	
	if (bioskey( 1) != 0)
		return; /* speedup */
	
	Pnames = GetResource("PNAMES");
	
	/* clear the box where the texture size will be drawn - see below */
	SetColor(DARKGRAY);
	DrawScreenBox( x0 - 179, y0 + 40, x0 - 82, y0 + 78);
	
	if (!(tinfo = FindTexture(texname)))
		return;
	
	/* read the info for this texture */
	xsize = tinfo->width;
	ysize = tinfo->height;
	scalex = tinfo->scalex ? tinfo->scalex : 8;
	scaley = tinfo->scaley ? tinfo->scaley : 8;

	SetColor( BLACK);
	DrawScreenText( x0 - 171, y0 + 40, "%dx%d", xsize, ysize);
	DrawScreenText (x0 - 179, y0 + 48, "(%dx%d)", (xsize << 3) / scalex, (ysize << 3) / scaley);
	
	DrawScreenText( x0 - 171, y0 + 68, "%d Patch%s", tinfo->nPatches, tinfo->nPatches==1?"":"es");
	
	if (bioskey( 1) != 0)
		return; /* speedup */
	
	/* not really necessary, except when xofs or yofs < 0 */
	++x1;
	if (x0 + xsize < x1)
	{
		x1 = x0 + xsize;
	}
	if (y0 + ysize < y1)
	{
		y1 = y0 + ysize;
	}
	GrSetClipBox (x0, y0, x1, y1);
	
	/* display the texture */
	pinfo = (PatchDesc *)((long)tinfo + sizeof(Texture));
	ForgetAfter = FALSE;
	for (n = 0; n < tinfo->nPatches; n++, pinfo++)
	{
		/* OK, now look up the pic's name in Pnames. */
		patchname = (long)Pnames + 4L + ((long)(pinfo->PatchNumber) * 8);
		
		strncpy(picname, (char *)patchname, 8);
		picname[ 8] = '\0';
		// [RH] preview texture with scaling
		DisplayPicFrac( x0 + (((long)pinfo->xOffset) << 3) / scalex, (y0*scaley)/8 + pinfo->yOffset,
				   x1, y1, strupr( picname), scalex<<13, scaley<<13);
	}
	
	/* de-allocate all patches 
	if (Caching == FALSE) {
	*/
		pinfo = (PatchDesc *)((long)tinfo + sizeof(Texture));
		for (n = 0; n < tinfo->nPatches; n++, pinfo++)
		{
			patchname = (long)Pnames + 4L + ((long)(pinfo->PatchNumber) * 8);
			strncpy(picname, (char *)patchname, 8);
			picname[ 8] = '\0';
			ForgetResource(picname);
		}

	/* restore the normal viewport */
	GrResetClipBox ();
}

// [RH] Named "Quick" because it's used by the fullscreen text browser--not because
// it's optimized for drawing textures quickly.
void DisplayWallQuick( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *texname, BCINT scale)
{
	int	n, xsize, ysize;
	Texture  *tinfo;
	PatchDesc *pinfo;
	char	 picname[9];
	long patchname;
	void *Pnames;
	long xstep, ystep, xscale, yscale, step;

	if (bioskey( 1) != 0 && Caching == FALSE)
		return; 

	if (texname[0] == '-') 
		return;

	x1 = x0 + (x1 - x0 + 1) / scale;
	y1 = y0 + (y1 - y0) / scale;
	scale = 1;

	Pnames = GetResource("PNAMES");

	if (!(tinfo = FindTexture(texname)))
	{
		SetColor(GRAY);
		DrawScreenLine(x0, y0, x1, y1 - 1);
		DrawScreenLine(x0, y1 - 1, x1, y0);
		return;
	}

	/* read the info for this texture */
	xsize = tinfo->width;
	ysize = tinfo->height;
	xscale = tinfo->scalex ? tinfo->scalex : 8;
	yscale = tinfo->scaley ? tinfo->scaley : 8;

	// [RH] Pick a scale appropriate for the cell
	xstep = (xsize<<19) / ((x1 - x0 + 1) * xscale);
	ystep = (ysize<<19) / ((y1 - y0) * yscale);
	step = max (65536, max (xstep, ystep));

	xstep = step * xscale / 8;
	ystep = step * yscale / 8;

	/* not really necessary, except when xofs or yofs < 0  */
	GrSetClipBox (x0, y0, x0 + (tinfo->width * 65536 / xstep), y0 + (tinfo->height * 65536 / ystep)); 

	/* display the texture */
	pinfo = (PatchDesc *)((long)tinfo + sizeof(Texture));
	ForgetAfter = FALSE;
	for (n = 0; n < tinfo->nPatches; n++, pinfo++)
	{
		patchname = (long)Pnames + 4L + ((long)(pinfo->PatchNumber) * 8);
		strncpy(picname, (char *)patchname, 8);
		picname[ 8] = '\0';
		DisplayPicFrac( x0 + ((long)(pinfo->xOffset) << 16) / xstep, (y0*ystep)/65536 + pinfo->yOffset,
				   x1, y1, strupr( picname), xstep, ystep);
	}

	/* de-allocate all patches  
	if (Caching == FALSE) { */
		pinfo = (PatchDesc *)((long)tinfo + sizeof(Texture));
		for (n = 0; n < tinfo->nPatches; n++, pinfo++)
		{
			patchname = (long)Pnames + 4L + ((long)(pinfo->PatchNumber) * 8);
			strncpy(picname, (char *)patchname, 8);
			picname[ 8] = '\0';
			ForgetResource(picname);
		}

	GrResetClipBox ();
}


/*
   Function to get the size of a wall texture
   It's a lot smaller now that we have functions to get a pointer
   to the texture structure automagically :-)
   */

void GetWallTextureSize( BCINT *xsize_r, BCINT *ysize_r, char *texname)
{
	Texture *t = FindTexture(texname);
	
	if(t) {
		*xsize_r = t->width;
		*ysize_r = t->height;
	}
	else {
		/* texture data not found */
		*xsize_r = -1;
		*ysize_r = -1;
	}
}




/*
   choose a floor or ceiling texture
   */

void ChooseFloorTexture( BCINT x0, BCINT y0, char *prompt, BCINT listsize, char **list, char *name)
{
	if (UseMouse)
		HideMousePointer(); 
	SwitchToVGA256();
	/* if we only have a 320x200x256 VGA driver, we must change x0 and y0.  Yuck! */
	if (GfxMode > -2) {
		x0 = -1;
		y0 = -1;
	}
	InputNameFromListWithFunc( x0, y0, prompt, listsize, list, 5, name, 64, 64, DisplayFloorTexture);
	SwitchToVGA16();
	if (UseMouse)
		ShowMousePointer();
}



/*
   choose a wall texture
   */

void ChooseWallTexture( BCINT x0, BCINT y0, char *prompt, BCINT listsize, char **list, char *name)
{
	if (UseMouse)
		HideMousePointer();
	SwitchToVGA256();
	/* if we only have a 320x200x256 VGA driver, we must change x0 and y0.  Yuck! */
	if (GfxMode > -2) {
		x0 = 0;
		y0 = -1;
	}
	InputNameFromListWithFunc( x0, y0, prompt, listsize, list,
		/* [RH] */ ScrMaxY < 480 ? 24 : 49, name, ScrMaxX<800 ? 256 : 512/* jff */,
		/* [RH] */ ScrMaxY < 480 ? 256 : 512, DisplayWallTexture);
	SwitchToVGA16();
	if (UseMouse)
		ShowMousePointer();
}



/*
   function used by qsort to sort the sprite names
   */
int SortSprites( const void *a, const void *b)
{
	return strcmp( *((char **)a), *((char **)b));
}



/*
   choose a "sprite"
   */

void ChooseSprite( BCINT x0, BCINT y0, char *prompt, char *sname)
{
	MDirPtr dir;
	BCINT n, listsize;
	char **list;
	char name[ 9];
	
	/* count the names */
	dir = FindMasterDir( MasterDir, "S_START");
	dir = dir->next;
	for (n = 0; dir && strcmp(dir->dir.name, "S_END"); n++) 
		dir = dir->next;
	listsize = n;
	/* get the actual names from master dir. */
	dir = FindMasterDir( MasterDir, "S_START"); 
	dir = dir->next;
	list = (char**) myalloc( listsize * sizeof( char *));
	for (n = 0; n < listsize; n++) {
		list[ n] = (char*) myalloc( 9 * sizeof( char));
		strncpy( list[ n], dir->dir.name, 8);
		list[ n][ 8] = '\0';
		dir = dir->next;
	}
	qsort( list, listsize, sizeof( char *), SortSprites);
	if (sname != NULL)
		strncpy( name, sname, 8);
	else
		strcpy( name, list[ 0]);
	/*   if (UseMouse)
		 HideMousePointer(); */
	SwitchToVGA256();
	/* if we only have a 320x200x256 VGA driver, we must change x0 and y0.  Yuck! */
	if (GfxMode > -2) {
		x0 = 0;
		y0 = -1;
	}
	ForgetAfter = TRUE;
	InputNameFromListWithFunc( x0, y0, prompt, listsize, list, 16, name, 320, 200, DisplaySprite);
	/*   SwitchToVGA16();
		 if (UseMouse)
		 ShowMousePointer(); */
	for (n = 0; n < listsize; n++)
		free( list[ n]);
	free( list);
}
