/*
** wadconv.c - zwadconv merged with ZETH to be more convenient
**
**---------------------------------------------------------------------------
** Copyright 1998-2002 Randy Heit
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
**
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
** IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
** IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
** INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
** NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
** THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**---------------------------------------------------------------------------
**
*/

#include "deu.h"
#include "levels.h"

extern unsigned char DoomTranslator[], HereticTranslator[];

#define PCD_TERMINATE			1
#define PCD_LSPEC1DIRECT		9
#define PCD_LSPEC2DIRECT		10
#define PCD_LSPEC3DIRECT		11
#define PCD_LSPEC4DIRECT		12
#define PCD_LSPEC5DIRECT		13

#define SIMPLE_HASTAG			0x80
#define SIMPLE_HASLINEID		0x40
#define SIMPLE_HAS2LINEIDS		0xc0

#define T_DESTINATION			14

// define names for the TriggerType field of the general linedefs

typedef enum
{
	WalkOnce,
	WalkMany,
	SwitchOnce,
	SwitchMany,
	GunOnce,
	GunMany,
	PushOnce,
	PushMany,
} triggertype_e;

typedef unsigned long ULONG;

static ULONG *FlyBehavior;
static ULONG FlySize, FlyScript;
static ULONG *ScriptStarts;
static char *FlySource, *FlyPt;

static BYTE *SectorTags;
static BYTE *RefCount;
static BYTE *UsedTags;

static BYTE *LineIDMaps;
static UBCINT *ReverseLineIDMaps;
static UBCINT *LineTags;

static Bool SomeProblems = FALSE;

static Bool UsedLineIDs;

static UBCINT freetag (UBCINT tag);
static BYTE maptag (UBCINT tag);
static UBCINT freelineid (UBCINT id);
static BYTE maplineid (UBCINT id);
static void put (const char *fmt, ...);

static Bool DoomTrans;

void InitConversion (void)
{
	DoomTrans = strnicmp (Game, "DOOM", 4) == 0;

	SectorTags = myalloc (65536);
	RefCount = myalloc(65536/8);
	UsedTags = myalloc(256/8);

	LineIDMaps = myalloc(65536);
	ReverseLineIDMaps = myalloc(256*2);
	LineTags = myalloc(32768*2);

	FlyBehavior = myalloc(2048*4);
	ScriptStarts = myalloc(255*4);
	FlyPt = FlySource = myalloc(65536);

	memset (SectorTags, 0, 65536);
	memset (RefCount, 0, 65536/8);
	memset (UsedTags, 0, 256/8);

	memset (LineIDMaps, 0, 65536);
	memset (ReverseLineIDMaps, 0, 256*2);
	memset (LineTags, 0, 32768*2);

	memset (FlyBehavior, 0, 2048*4);
	memset (ScriptStarts, 0, 255*4);
	*FlyPt = 0;

	UsedLineIDs = FALSE;
	SomeProblems = FALSE;
	FlySize = 2;
	FlyScript = 0;
}

void FinishConversion (void)
{
	/* An empty behavior lump containing no scripts or strings */
	static const BYTE NullBehavior[16] = "ACS\0\x8\0\0\0\0\0\0\0\0\0\0\0";
	static const char NullScripts[] = "#include \"zcommon.acs\"\n\n";

	int i;

	if (UsedLineIDs)
	{
		// Make sure that everything that could be referenced by a
		// line id is. We also need an open script to set the lines
		// up if we can't just give them a special that sets their ID.
		LDPtr line;
		//Bool gotscript = FALSE;

		put ("Setting LineIDs\n");
		for (i = 0, line = LineDefs; i < NumLineDefs; ++i, ++line)
		{
			if (LineIDMaps[LineTags[i]])
			{
				if (line->special != 215 /*Teleport_Line*/ &&
					line->special != 208 /*TranslucentLine*/ &&
					line->special != 222 /*Scroll_Texture_Model*/)
				{
					if (line->special == 0)
					{
						// Set this line to Line_SetIdentification
						line->special = 121; /*Line_SetIdentification*/
						line->arg[0] = LineIDMaps[LineTags[i]];
					}
					else
					{
						// This line does something special! Gak!
						put ("FIX THIS: Linedef %d (might) need id %d but already has a special.\n",
							 i, LineIDMaps[LineTags[i]]);
						SomeProblems = TRUE;
					}
				}
			}
		}
	}

	if (FlyScript)
	{
		// We generated a behavior lump, so let ZETH know about it.
		FlyBehavior[1] = FlySize * 4;
		FlyBehavior[FlySize++] = FlyScript;
		for (i = 0; i < FlyScript; i++, FlySize += 3)
		{
			FlyBehavior[FlySize+0] = i + 1;
			FlyBehavior[FlySize+1] = ScriptStarts[i];
			FlyBehavior[FlySize+2] = 0;
		}
		FlyBehavior[FlySize++] = 0;		// 0 strings

		BehaviorSize = FlySize * 4;
		Behavior = myalloc (BehaviorSize);
		memcpy (Behavior, FlyBehavior, BehaviorSize);
		put ("%d script%s created\n", FlyScript, FlyScript > 1 ? "s" : "");

		ScriptsSize = FlyPt - FlySource;
		Scripts = myalloc (ScriptsSize);
		memcpy (Scripts, FlySource, ScriptsSize);
	}
	else
	{
		BehaviorSize = 16;
		Behavior = myalloc (16);
		memcpy (Behavior, NullBehavior, 16);

		ScriptsSize = sizeof(NullScripts) - 1;
		Scripts = myalloc (ScriptsSize);
		memcpy (Scripts, NullScripts, ScriptsSize);
	}

	// Set the TID of all teleport destinations to match the tag
	// of their containing sectors.
	for (i = 0; i < NumThings; ++i)
	{
		if (Things[i].type == T_DESTINATION)
		{
			BCINT secnum = GetCurObject (OBJ_SECTORS, Things[i].xpos, Things[i].ypos, 0);
			if (secnum != -1)
			{
				Things[i].tid = Sectors[secnum].tag;
			}
		}
	}

	free (FlySource);
	free (ScriptStarts);
	free (FlyBehavior);
	free (LineTags);
	free (ReverseLineIDMaps);
	free (LineIDMaps);
	free (UsedTags);
	free (RefCount);
	free (SectorTags);
}

// Returns TRUE if a 666 or 667 tag was referenced, and there
// were no free scripts.
Bool TranslateLineDef (int linenum, int special, int tag)
{
	const char *argspec;
	int numparms;
	BYTE *tlate;
	LDPtr ld = &LineDefs[linenum];
	BCINT flags = ld->flags;
	Bool passthrough;
	Bool usedtag = FALSE;
	int i;

	if (DoomTrans)
	{
		tlate = DoomTranslator;
	}
	else
	{
		tlate = HereticTranslator;
	}

	passthrough = (flags & ML_PASSUSE_BOOM);

	flags = flags & 0x01ff;	// Ignore flags unknown to DOOM

	LineTags[linenum] = tag;

	// 0 specials are never translated.
	if (special == 0)
	{
		ld->special = 0;
		ld->flags = flags;
		memset (ld->arg, 0, sizeof(ld->arg));
		return FALSE;
	}

	// Check if this is a regular linetype
	if (tlate[0] == 'N' && tlate[1] == 'O' && tlate[2] == 'R' && tlate[3] == 'M')
	{
		int count = (tlate[4] << 8) | tlate[5];
		tlate += 6;
		while (count)
		{
			int low = (tlate[0] << 8) | tlate[1];
			int high = (tlate[2] << 8) | tlate[3];
			tlate += 4;
			if (special >= low && special <= high)
			{ // found it, so use the LUT
				BYTE *specialmap = tlate + (special - low) * 7;

				flags = flags | ((specialmap[0] & 0x3f) << 9);

				if (passthrough && (GET_SPAC(flags) == SPAC_USE))
				{
					flags &= ~ML_SPAC_MASK;
					flags |= SPAC_USETHROUGH << ML_SPAC_SHIFT;
				}
				ld->special = specialmap[1];
				ld->arg[0] = specialmap[2];
				ld->arg[1] = specialmap[3];
				ld->arg[2] = specialmap[4];
				ld->arg[3] = specialmap[5];
				ld->arg[4] = specialmap[6];
				switch (specialmap[0] & 0xc0)
				{
				case SIMPLE_HASTAG:
					usedtag = TRUE;
					if (tag != 666 && tag != 667)
						tag = maptag (tag);
					ld->arg[0] = tag;
					break;

				case SIMPLE_HASLINEID:
					UsedLineIDs = TRUE;
					tag = maplineid (tag);
					ld->arg[0] = tag;
					break;

				case SIMPLE_HAS2LINEIDS:
					UsedLineIDs = TRUE;
					tag = maplineid (tag);
					ld->arg[0] = ld->arg[1] = tag;
					break;
				}
				goto finishline;
			}
			tlate += (high - low + 1) * 7;
			count--;
		}
	}

	// Check if this is a BOOM generalized linetype
	if (tlate[0] == 'B' && tlate[1] == 'O' && tlate[2] == 'O' && tlate[3] == 'M')
	{
		int count = (tlate[4] << 8) | tlate[5];
		tlate += 6;

		// BOOM translators are stored on disk as:
		//
		// WORD <first linetype in range>
		// WORD <last linetype in range>
		// BYTE <new special>
		// repeat [BYTE op BYTES parms]
		//
		// op consists of some bits:
		//
		// 76543210
		// ||||||++-- Dest is arg[(op&3)+1] (arg[0] is always tag)
		// |||||+---- 0 = store, 1 = or with existing value
		// ||||+----- 0 = this is normal, 1 = x-op in next byte
		// ++++------ # of elements in list, or 0 to always use a constant value
		//
		// If a constant value is used, parms is a single byte containing that value.
		// Otherwise, parms has the format:
		//
		// WORD <value to AND with linetype>
		// repeat [WORD <if result is this> BYTE <use this>]
		//
		// These x-ops are defined:
		//
		// 0 = end of this BOOM translator
		// 1 = dest is flags

		while (count)
		{
			int low = (tlate[0] << 8) | tlate[1];
			int high = (tlate[2] << 8) | tlate[3];
			short oflags = flags;
			tlate += 4;


			// Assume we found it and translate
			switch (special & 0x0007)
			{
			case WalkMany:
				flags |= ML_REPEAT_SPECIAL;
			case WalkOnce:
				flags |= SPAC_CROSS << ML_SPAC_SHIFT;
				break;

			case SwitchMany:
			case PushMany:
				flags |= ML_REPEAT_SPECIAL;
			case SwitchOnce:
			case PushOnce:
				if (passthrough)
					flags |= SPAC_USETHROUGH << ML_SPAC_SHIFT;
				else
					flags |= SPAC_USE << ML_SPAC_SHIFT;
				break;

			case GunMany:
				flags |= ML_REPEAT_SPECIAL;
			case GunOnce:
				flags |= SPAC_IMPACT << ML_SPAC_SHIFT;
				break;
			}

			// We treat push triggers like switch triggers with zero tags.
			usedtag = (special & 7) != PushMany && (special & 7) != PushOnce;
			ld->arg[0] = usedtag ? (tag = maptag(tag)) : 0;
			ld->arg[1] = ld->arg[2] = ld->arg[3] = ld->arg[4] = 0;

			ld->special = *tlate++;
			for (;;)
			{
				UBCINT *destp;
				UBCINT flagtemp;
				BYTE op = *tlate++;
				BYTE dest;
				BYTE val = 0;	// quiet, GCC
				Bool found;
				int lsize;

				dest = op & 3;
				if (op & 8)
				{
					BYTE xop = *tlate++;
					if (xop == 0)
						break;
					else if (xop == 1)
						dest = 4;
				}
				if (dest < 4)
				{
					destp = (UBCINT *)&ld->arg[dest+1];
				}
				else
				{
					flagtemp = (flags >> 9) & 0x3f;
					destp = &flagtemp;
				}
				lsize = op >> 4;
				if (lsize == 0)
				{
					val = *tlate++;
					found = TRUE;
				}
				else
				{
					UBCINT mask = (tlate[0] << 8) | tlate[1];
					tlate += 2;
					found = FALSE;
					for (i = 0; i < lsize; i++)
					{
						UBCINT filter = (tlate[0] << 8) | tlate[1];
						if ((special & mask) == filter)
						{
							val = tlate[2];
							found = TRUE;
						}
						tlate += 3;
					}
				}
				if (found)
				{
					if (op & 4)
					{
						*destp |= val;
					}
					else
					{
						*destp = val;
					}
					if (dest == 4)
					{
						flags = (flags & ~0x7e00) | (flagtemp << 9);
					}
				}
			}
			if (special >= low && special <= high)
			{ // Really found it, so we're done
				goto finishline;
			}

			flags = oflags;
			count--;
		}
	}

	// Don't know what to do, so 0 it
	ld->special = 0;
	ld->flags = flags;
	memset (ld->arg, 0, sizeof(ld->arg));
	return FALSE;

finishline:
	ld->flags = flags;

	argspec = GetLineDefArgs (ld->special);
	for (numparms = 5; numparms > 0; --numparms)
	{
		if (argspec[numparms-1] != '-')
			break;
	}

	if (usedtag && (tag == 666 || tag == 667) && numparms)
	{
		// Sector tags 666 and 667 are special cases:
		// We need to preserve them, so we construct a script for them.
		// If the special takes no parameters, though, we don't need to
		// bother, since the tag isn't actually used by it in that case.
		if (FlyScript >= 255)
		{
			return TRUE;
		}
		else
		{
			char *name = GetLineDefTypeLongName (ld->special), *namespc;

			ScriptStarts[FlyScript] = FlySize * 4;
			FlyBehavior[FlySize++] = PCD_LSPEC1DIRECT + numparms - 1;
			FlyBehavior[FlySize++] = ld->special;
			FlyBehavior[FlySize++] = tag;
			for (i = 1; i < numparms; i++)
			{
				FlyBehavior[FlySize++] = ld->arg[i];
			}
			FlyBehavior[FlySize++] = PCD_TERMINATE;

			if (FlyScript == 0)
			{
				FlyPt += sprintf (FlyPt, "#include \"zcommon.acs\"\n");
			}
			namespc = strchr (name, ' ');
			if (namespc)
				name = namespc + 1;
			FlyPt += sprintf (FlyPt, "\nscript %lu (void)\n{\n%s (%d",
				FlyScript+1, name, tag);
			for (i = 1; i < numparms; ++i)
			{
				FlyPt += sprintf (FlyPt, ", %d", ld->arg[i]);
			}
			FlyPt += sprintf (FlyPt, ");\n}\n");
			FlyScript++;

			ld->special = 226; /*ACS_ExecuteAlways*/
			ld->arg[0] = (BYTE)FlyScript;
			ld->arg[1] = ld->arg[2] = ld->arg[3] = ld->arg[4] = 0;

			return FALSE;
		}
	}
	return FALSE;
}

void ConvertSectors (void)
{
	int i;

	for (i = 0; i < NumSectors; ++i)
	{
		SPtr sector = &Sectors[i];
		if (SectorTags[sector->tag])
		{
			sector->tag = SectorTags[sector->tag];
		}
		else if ((RefCount[sector->tag/8] & (1 << (sector->tag & 7))) == 0 &&
				 sector->tag && sector->tag != 666 && sector->tag != 667)
		{
			//put ("Sector tag %d is unreferenced\n", sector->tag);
			//RefCount[sector->tag/8] |= 1 << (sector->tag & 7);
		}
		sector->special = TranslateSectorType (sector->special);
	}
}

int TranslateSectorType (int special)
{
	int high;

	if (special == 0)
	{
		// Don't translate type 0 to 64
		return 0;
	}
	else if (special == 9)
	{
		return 0x0400;
	}
	else if (DoomTrans)
	{
		// This supports phased lighting with specials 21-24
		high = (special & 0xfe0) << 3;
		special &= 0x1f;
		if (special < 21)
		{
			return high | (special + 64);
		}
		else if (special < 40)
		{
			return high | (special - 20);
		}
	}
	else
	{
		high = (special & 0xfc0) << 3;
		special &= 0x3f;
		if (special == 5)
		{
			return high | 82;
		}
		else if (special == 16)
		{
			return high | 83;
		}
		else if (special == 4)
		{
			return high | 84;
		}
		else if (special < 20)
		{
			return high | (special + 64);
		}
		else if (special < 40)
		{
			return high | (special + 205);
		}
	}
	return high | special;
}

static UBCINT freetag (UBCINT tag)
{
	if (tag < 256 && ((UsedTags[tag/8] & (1 << (tag & 7))) == 0))
	{
		return tag;
	}
	else
	{
		int i;

		if (tag == 666 || tag == 667)
			put ("WARNING! ");

		for (i = 128; i < 256; i++)
		{
			if ((UsedTags[i/8] & (1 << (i&7)))== 0)
			{
				put ("Tag %d remapped to %d\n", tag, i);
				return i;
			}
		}
		for (i = 127; i >= 0; i--)
		{
			if ((UsedTags[i/8] & (1 << (i&7))) == 0)
			{
				put ("Tag %d remapped to %d\n", tag, i);
				return i;
			}
		}
	}
	put ("FIX THIS: Could not remap tag %d (using 0)!\n", tag);
	SomeProblems = TRUE;
	return 0;
}


static BYTE maptag (UBCINT tag)
{
	if (tag == 0)
	{
		return 0;
	}
	else if (SectorTags[tag] == 0)
	{
		// Try to find a good tag for this one (preferably the same)
		int newtag = freetag (tag);

		SectorTags[tag] = newtag;
		UsedTags[newtag/8] |= 1 << (newtag & 7);
	}

	return SectorTags[tag];
}

static UBCINT freelineid (UBCINT id)
{
	if (id < 256 && !ReverseLineIDMaps[id])
	{
		return id;
	}
	else
	{
		int i;

		for (i = 0; i < 255; i++)
			if (ReverseLineIDMaps[i] == 0)
			{
				put ("LineID %d remapped to %d\n", id, i);
				return i;
			}
	}
	put ("FIXTHIS: Could not remap lineid %d (using 0)!\n", id);
	SomeProblems = TRUE;
	return 0;
}


static BYTE maplineid (UBCINT id)
{
	if (id == 0)
	{
		return 0;
	}
	else if (LineIDMaps[id] == 0)
	{
		// Try to find a good lineid for this one (preferably the same)
		int newid = freelineid (id);

		LineIDMaps[id] = newid;
		ReverseLineIDMaps[newid] = id;
	}

	return LineIDMaps[id];
}

static void put (const char *fmt, ...)
{
#if 0
	va_list argptr;

	va_start (argptr, fmt);
	vprintf (fmt, argptr);
	vfprintf (logfile, fmt, argptr);
	va_end (argptr);
#endif
}
